\ PicForth example program.

\ This program buffers incoming bytes from the serial port and writes them to
\ an LCD display.

\ Conditional compilation flag.
0 constant simulation
0 constant test_serial

fload picf16.f
fload math16.f
fload interupt.f
fload misc.f
fload displayl.f
fload disp.f
fload serial.f



\ ---------  Serial Receive Interrupt Handling --------------------

\ Size of the serial buffer.
25 constant ser_buf_len

\ Declare the serial buffer.
ser_buf_len array serial_buffer

\ The value of the end of the serial buffer. 
serial_buffer ser_buf_len + constant ser_buf_end

\ Two pointers into the buffer to make a cicular queue.
variable get_ptr
variable put_ptr

\ Some initialisation details.
code enable_ser_buf
        periphbank BANK1
        bsf PIE, RCIE           ; enable RX interrupts
        periphbank BANK3
        bsf INTSTA, PEIE        ; enable peripheral interrupts
        call _enable_ints       ; enable global interrupts
        return
endcode

: reset_ser_buf  
        serial_buffer dup get_ptr ! put_ptr !
        ;
        
: inc_ptr ( ptr_addr -- )
        dup @ 1+ 
        dup ser_buf_end >= if drop serial_buffer then  \ Wrap pointer.
        swap ! ; 

\ Buffer a serial reception, called by the interrupt routine.
: buf_RX
        RX dup -1 = if
            drop 
        else 
            put_ptr @ c!
            put_ptr inc_ptr
        then ;

\ Flag if a char is available in the buffer.
: RX_avail?  ( -- f)
           get_ptr @ put_ptr @ <> ;
        
        
\ Blocking read from the buffer.        
: get_RX ( -- ch )
        begin
           RX_avail?
        until
        
        get_ptr @ c@
        get_ptr inc_ptr
        ;

\ ---------  Peripheral Interrupt Handling --------------------

32 :interrupt (peripheral_interrupt)

[code]
; vector to the appropriate interrupt handler.
        dup                     ; make space on the stack
        periphbank BANK1
        movpf PIR, TLOW         ; store PIR in TLOW
        movpf PIE, WREG         
        andwf TLOW, F           ; PIR AND PIE => TLOW
        btfsc TLOW, RCIF        ; check for RX interrupt
        call _buf_RX
        drop                    ; return space to the stack
endcode

interrupt;


\ ---------  Main Program -------------------------------------

\ Notify us if we can't keep up with incoming characters.
: ?RX_ovf
      RX_overflow? if 
         page ." RX Overflow!" 500 ms
         rx_disable rx_enable   \ Reset the serial port.
      then ;

\ Echo any received characters to the LCD screen.
: ?inchar
           RX_avail? 0= if exit then
           
           get_RX
           
           \ Debug the incoming serial bytes.
           [ test_serial ] [if] . space exit [then]
           
           emit 

           ?RX_ovf
           ;

: main

        \ Initialize peripheral units.
        1000 MS      \ Power up delay.
        LCD_INSTRUCTION_RESET
        scroll_mode
        hide_cursor
 
        \ Show a blurb on startup.
        page ." PicForth example" CR
             ." program."
             
        3000 ms page
        
        
        \ Initialise the serial buffer.        
        reset_ser_buf
        enable_ser_buf  \ Enables peripheral interrupts.
        
        \ Enable the serial port at 9600 baud.
        25 setbrg  \ 9600 baud at 16 MHz
        serial_enable
        
        \ Enable reception.
        rx_enable_parity
        rx_enable
        

        \ Infinite loop.
        begin
           ?inchar  \ Process incoming characters.
        again
;        

