.( 
        blcd8.f
        Hitachi HD44780 controller LCD display routines.
)


\ This code is Copyright 1998, Peter Csapo. It is has been distributed free
\ of charge, and no claim is made to its accuracy, reliability or performance
\ in any way. 


\ Control words for the liquid crystal display.
\ Hitachi controller HD44780.
\ This code uses the 8 bit data bus option.
\ Connected to port D.
\ Assume PIC speed is 20 MHz.
\ Pin assignments are: RD0..7  -- DB0..7  data bus
\                      RE1 - R/W 1:read                0:write
\                      RE2 - RS  1:data instruction    0:control instruction
\                      RE0 - enable
\
\ Original: May 1, 1997
\ By Peter Csapo
\ ===============================================================
\ Modified May 6, 1997
\   -- Better documentation on the controller in the AND databook
\      so a lot has changed.
\ June 12, 1997
\   -- Renamed LCD_READ to LCD@
\ June 25, 1997
\   -- Adjusted for 20 chars by 4 line display.
\ ===============================================================
\ Assumptions:
\       -- PORTE is set to be output on reset, and is assumed to never be
\          changed.

code LCD_PULSE
        bsf PORTE, 0
        goto $+1                ; 2 nops
        bcf PORTE, 0
        return
endcode

code LCD_WAIT
; Wait for the busy bit to go low.
; Assume bank is correct.
        ; Make the data bits inputs.
        setf DDRD, F            ; set port d to input
_LCD_WAIT_LOOP
        ; Set for reading busy flag.
        clrf PORTE, F
        bsf PORTE, 1            ; set R/W bit for read
        ; Put the flag into WREG.
        bsf PORTE, 0            ; enable LCD read
        nop
        clrf WREG, F
        btfsc PORTD, 7          ; test busy flag
        setf WREG, F
        clrf PORTE, F           ; end pulse

        ; Repeat if busy bit is set.
        tstfsz WREG
        goto _LCD_WAIT_LOOP
        return
endcode



code LCD_CMD
; Write TLOW to the LCD as a command
        periphbank BANK1
        call _LCD_WAIT          ; wait till busy bit goes low
        clrf DDRD, F            ; set RD0..7 for output
        clrf PORTE, F           ; set control lines for a command write
        movpf TLOW, PORTD       ; bus the data
        call _LCD_PULSE         ; write the data to the LCD
        call _drop              ; drop the command byte
        return
endcode

code LCD_ON
        ; display on
        movlw 0x0C
        call _push_byte_literal
        call _LCD_CMD
        return
endcode

code LCD_OFF
        ; display off
        movlw 0x08
        call _push_byte_literal
        call _LCD_CMD
        return
endcode

code LCD_POWER_ON_RESET
; Power on reset for the LCD display
        periphbank BANK1
        clrf PORTD, F           ; clear RD0..RD7
        clrf DDRD, F            ; set RD1..RD7 for output
        clrf PORTE, F           ; clear RS R/W and enable
        clrf DDRE, F            ; set port E for output
        
        ; add a delay to let the LCD initialize itself
        setf WREG, F            ; push 255
        call _push_byte_literal
        call _MS                ; call MS function to delay

        ; wait for the busy flag to drop
        call _LCD_WAIT
        
        movlw 0x38              ; function set 2 lines
        call _push_byte_literal
        call _LCD_CMD

        ; turn the display on                ;
        call _LCD_ON

        ; set DD RAM address to top left
        movlw 0x80
        call _push_byte_literal
        call _LCD_CMD
        
        return
endcode

code LCD_INSTRUCTION_RESET
; during this reset operation.
; Send the initialization sequence to the LCD controller.
; On powerup, a 15 MS delay must be added after the voltage
; reaches 4.5V before this function is called (by spec).
; In practise, use a longer delay to be sure.
        periphbank BANK1
        clrf PORTD, F           ; clear RD0..RD7
        clrf DDRD, F            ; set RD1..RD7 for output
        clrf PORTE, F           ; clear RS R/W and enable
        clrf DDRE, F            ; set port E for output
        movlw 0x30              ; function set
        movwf PORTD
        call _LCD_PULSE         ; write to the LCD
        ; delay 5 ms
        movlw 0x5
        call _push_byte_literal
        call _MS                ;
        call _LCD_PULSE         ; write again to the LCD
        ; delay more than 100 ms
        movlw 0x80
        call _push_byte_literal
        call _MS
        call _LCD_PULSE         ; write a third time to the LCD

        ; now we can communicate normally, checking busy flag
        movlw 0x38              ; function set 2 lines
        call _push_byte_literal
        call _LCD_CMD

        call _LCD_ON            ; turn on display
        call _LCD_page          ; clear display

        ; set for increment / no shift
        movlw 0x06
        call _push_byte_literal
        call _LCD_CMD

        ; set DD RAM address to top left
        movlw 0x80
        call _push_byte_literal
        call _LCD_CMD

        return
endcode

code LCD@
; ( -- C)
; Read a byte from the LCD memory.
        periphbank BANK1
        call _LCD_WAIT
        setf DDRD, F            ; set D for input
        clrf PORTE, F
        bsf PORTE, 1
        bsf PORTE, 2
        nop
        bsf PORTE, 0            ; enable LCD
        call _pushtop
        clrf THIGH, F
        movfp PORTD, TLOW       ; read LCD
        clrf PORTE, F
        return
endcode


code LCD!
; ( c -- )
; Write TLOW to the LCD as data
        periphbank BANK1
        call _LCD_WAIT          ; wait till busy bit goes low
        clrf DDRD, F            ; set RD0..7 for output
        clrf PORTE, F
        bsf PORTE, 2            ; set control lines for write data
        movfp TLOW, PORTD       ; bus the data
        call _LCD_PULSE         ; write to the LCD
        call _drop              ; drop the byte
        return
endcode


0  constant LCD_ROW0_BASE
64 constant LCD_ROW1_BASE
20 constant LCD_ROW2_BASE
84 constant LCD_ROW3_BASE

\ Updated to use a 4 line display.
: LCD_at-xy  ( x y --)
        dup LAST_LINE u> if drop drop exit then
        
        dup 0= if 
           LCD_ROW0_BASE
        else dup 1 = if 
           LCD_ROW1_BASE
        else dup 2 = if 
           LCD_ROW2_BASE
        else
           LCD_ROW3_BASE 
        then then then
        nip
         + 128 or
        LCD_CMD
        ;
                
\ Using a 20 char display.
: LCD_scroll_up

        \ Copy the n+1 line onto the nth.
        LAST_LINE 0 do
           DISPLAY_WIDTH 0 do
              I J 1 + LCD_at-xy LCD@
              I J LCD_at-xy LCD!
           loop
        loop

        \ Clear the bottom line.
        0 LAST_LINE LCD_at-xy
        DISPLAY_WIDTH 0 do
           BL LCD!
        loop
        ;


: LCD_CR
        0 YPOS@ 1+
        dup LAST_LINE u> if
           drop
           SCRNmode@ SCRN_SCROLL = 
           if 
              LCD_scroll_up 
              LAST_LINE 
           else 
              0 
           then 
        then
        LCD_at-xy
        ;

1 constant LCD_PAGE_CMD
14 constant LCD_SHOWCURSOR_CMD
12 constant LCD_HIDECURSOR_CMD

: LCD_page
        LCD_PAGE_CMD LCD_CMD ;


: LCD_show_cursor
        LCD_SHOWCURSOR_CMD LCD_CMD ;

: LCD_hide_cursor
        LCD_HIDECURSOR_CMD LCD_CMD ;

