
.(
        I2C.f
        I2C master software interface
)



\ This code is Copyright 1998, Peter Csapo. It is has been distributed free
\ of charge, and no claim is made to its accuracy, reliability or performance
\ in any way. 




\  Control functions to send and recieve data on an I2C bus.
\        for interfacing with serial eeproms.

\ RB2 line will function as SCL on the I2C bus
\ RB3 line will be the SDA

\ PIC1744 at 16 MHz gives an instruction time of 250ns


code I2C_INIT
        #define SCL     2
        #define SDA     3
        periphbank BANK0
        ; call _RB_pullup_off   ; turn off pullups on portB
        clrf PORTB, F           ; clear PORTB
        bcf DDRB, SCL           ; put SCL, SDA line in  output state
        bcf DDRB, SDA
        return
endcode

:zzz I2C_ERROR  ( n --) 
        CR P." I2C ERROR " .
        begin again   \ Never return.
        ;

code I2C_err
        ; put the error code on the stack and goto EEPROM_ERROR
        movwf TLOW
        clrf THIGH, F
        goto _I2C_ERROR
endcode
unremember

code delay_4000_ns
        ; delay for 16 instructions
        ; 2 for call + 2 for return need 12 more
        clrf WREG, F    
        bsf WREG, 2
        ; loop 4 times with 3 instruction times = 12 cycles
__delay_4000_loop
        decfsz WREG
        goto __delay_4000_loop
        return          ; more than enough
endcode

code delay_4_ops
        return
endcode

code BSTART
        ;
        ; Send start bit to I2C device
        ;
        periphbank BANK0                ; address portb
        clrf PORTB, F
        bsf PORTB, SDA                  ; make sure SDA is high
        bcf DDRB, SDA                   ; output SDA
        bsf PORTB, SCL                  ; set clock high
        bcf DDRB, SCL                   ; output SCL
        call _delay_4_ops
        movlw 0x1                       ; ready error code 1
        btfss PORTB, SCL                ; SCL locked low?
        call _I2C_err
        call _delay_4000_ns             ; setup time is 4700 ns
        bcf PORTB, SDA                  ; SDA goes low during SCL high
        call _delay_4000_ns             ; hold time is 4000 ns
        bcf PORTB, SCL                  ; set clock low
        call _delay_4000_ns
        return
endcode

code BSTOP
        ;
        ; Send stop bit to I2C device
        ;
        periphbank BANK0                ; address PORTB
        clrf PORTB, F
        bcf DDRB, SCL                   ; put SCL, SDA line in  output state
        bcf PORTB, SDA                  ; return SDA low
        bcf DDRB, SDA
        call _delay_4_ops
        bsf PORTB, SCL                  ; set clock high
        call _delay_4_ops
        movlw 0x2                       ; ready error code
        btfss PORTB, SCL                ; clock not high?
        call _I2C_err
        call _delay_4000_ns             ; stop setup time: 4000 ns
        bsf PORTB, SDA                  ; sda goes low to high during SCL high
        call _delay_4_ops
        movlw 0x3                       ; ready error code
        btfss PORTB, SDA                ; locked low ?
        call _I2C_err
        call _delay_4000_ns
        ; bcf PORTB, SCL                  ; drive the clock low
        ; call _delay_4000_ns
        return
endcode

code RXBIT
        ; fetch a bit from the serial eeprom
        ; puts bit fetched in least significant bit of TLOW
        ;
        ;
        periphbank BANK0                ; address PORTB
        clrf PORTB, F
        bcf DDRB, SCL                   ; 
        bsf DDRB, SDA                   ; set SDA for input, SCL output
        bsf PORTB, SCL                  ; set clock high
_RX_wait_state                                        
        call _delay_4_ops
        btfss PORTB, SCL                ; clock high?
        goto _RX_wait_state             ; slave is clock stretching
        bcf TLOW, 0                     ;
        btfsc PORTB, SDA                ; read SDA
        bsf TLOW, 0                     ;
                                        ; minimum high time is 4000 ns
        call _delay_4000_ns
        bcf PORTB, SCL                  ; set SCL low
        call _delay_4000_ns
        return
endcode


\ Changed Oct. 25, 1997.
\ Now uses WREG.
code TXBIT
        ; store a bit to the serial eeprom
        ; writes the most significant bit of WREG
        ;
        periphbank BANK0                ; address PORTB
        clrf PORTB, F
        bcf DDRB, SCL                    ; put SCL, SDA line in  output state
        bcf DDRB, SDA
        btfss WREG, 7
        goto _bitstore_0
        bsf PORTB, SDA                  ; set SDA high
        movlw 0x5                       ; setup error code
        call _delay_4_ops
        btfss PORTB, SDA                ; not set high ?
        call _I2C_err
        goto _bitstore_clk1
_bitstore_0
        bcf PORTB, SDA
        nop
_bitstore_clk1
        bsf PORTB, SCL                  ; set clock high
        goto $ + 1                      ; 2 nops
        movlw 0x6                       ; setup error code 6
        btfss PORTB, SCL
        call _I2C_err
        call _delay_4000_ns
        bcf PORTB, SCL                  ; return clock to low
        call _delay_4000_ns
        return
endcode

code RXBYTE
;   ( -- x1 )    \ Receive a byte from the serial eeprom.
        call _false             ; create space on stack
; use high byte for count
        bsf THIGH, 3            ; count of 8
__RXBYTELOOP
        rlncf TLOW, F            ;  shift left
        call _RXBIT
        decfsz THIGH, F
        goto __RXBYTELOOP
        return
endcode

\ Acknowledge a received byte.
code ACK
        clrf WREG, F    ; 0 is an ack
        goto _TXBIT
endcode

\ Don't acknowledge a received byte.
code NACK
        setf WREG, F    ; 1 is a nack
        goto _TXBIT
endcode

code TXBYTE
;   ( x1 -- ack) \ Sends the high byte of x1 to the serial eeprom.
;                 \ Gets the acknowledge.

; use the low byte to count
        clrf TLOW, F
        bsf TLOW, 3             ; put 8 in tlow for count
__TXBYTELOOP
        movfp THIGH, WREG       ; put the byte to send in WREG
        call _TXBIT        
        rlncf THIGH, F
        decfsz TLOW, F          ; decrement the count 
        goto __TXBYTELOOP       ; loop until count is 0
        
        ; TLOW is zeroed.        
        clrf THIGH, F           ; zero top cell
        goto _RXBIT             ; get ack
        ; defer return
endcode

