;******************************************************************************
; Internal FORTH procedures: not accessible as FORTH words.
;******************************************************************************

; This code is Copyright 1998, Peter Csapo. It is has been distributed free
; of charge, and no claim is made to its accuracy, reliability or performance
; in any way. 


;***************************************************
;!### CODE  Word: pushtop
_pushtop
        dup
        return
;******************************************************************************
;***************************************************
;!### CODE  Word: push_byte_literal
_push_byte_literal
        ; Push the literal byte in W onto the stack
        ; setting the high byte to zero.
        decf SP, F
        movfp TLOW, STACK
        decf SP, F
        movfp THIGH, STACK
        movwf TLOW
        clrf THIGH, F
        return
;******************************************************************************    
; MACRO  pushaddr
pushaddr macro  lit_word
        movlw low lit_word
        call _push_byte_literal
        movlw high lit_word
        movwf THIGH
        endm
;******************************************************************************    
;***************************************************
;!### CODE  Word: movwf_thigh
_movwf_thigh
        ; Put wreg into THIGH
        movwf THIGH
        return
;******************************************************************************
;***************************************************
;!### CODE  Word: poptop
_poptop
        drop
        return
;******************************************************************************
;***************************************************
;!### CODE  Word: 2rdrop
_2rdrop
        decf RSP, F
        decf RSP, F
        decf RSP, F
        decf RSP, F
        return
;******************************************************************************
;***************************************************
;!### CODE  Word: if_internal
_if_internal
        ; internal code in an if word
        ; process the flag
        ; sets WREG if the top stack item is zero, else clears WREG
        movfp TLOW, WREG
        iorwf THIGH, W
        drop
        tstfsz WREG
        retlw  0x00             ; was not zero
        retlw  0xFF             ; was zero
;******************************************************************************
;***************************************************
;!### CODE  Word: limit_reached
_limit_reached
        ; internal code in the loop word
        ; increments the index and compares it to the limit
        ; index and limit will be removed from the stack if equal
        ; returns 0 in wreg if equal, else returns  -1
        movpf RSTACK, XLOW             ; move the index into X
        decf RSP, F
        movpf RSTACK, XHIGH
        decf RSP, F
        infsnz XLOW, F                  ; increment the index
        incf XHIGH, F
        movfp RSTACK, WREG              ; compare the low bytes
        cpfseq XLOW
        goto _limit_neq_B
        decf RSP, F
        movfp RSTACK, WREG              ; compare the high bytes
        cpfseq XHIGH
        goto _limit_neq_A
        decf RSP, F                     ; equal, knock the last byte off
        retlw 0x00                      ; the stack and return 0
_limit_neq_A
        incf RSP, F                     ; restore the high byte of the limit
        movfp XLOW, RSTACK
_limit_neq_B                            ; restore the index to the return stack
        incf RSP, F
        movfp XHIGH, RSTACK
        incf RSP, F
        movfp XLOW, RSTACK
        retlw 0xFF                      ; return -1
;********************************************************************************
;***************************************************
;CODE  Word: _rcall_return
;__rcall_return
;        movpf RSTACK, WREG            ; get low byte return address
;        decf RSP, F                     
;        movfp RSTACK, PCLATH          ; restore high byte return address
;        decf RSP, F
;        movwf PCL                     ; return
;*********************************************************************************
;***************************************************
;!### CODE  Word: _settop
__settop
        setf THIGH, F
        setf TLOW, F
        return
;*********************************************************************************
;***************************************************
;!### CODE  Word: _clrtop
__clrtop
        clrf TLOW, F
        clrf THIGH, F
        return
;*********************************************************************************
;***************************************************
;!### CODE  Word: _rcall_prologue
__rcall_prologue
        ; Put the return index in wreg onto the return stack.
        incf RSP, F
        movwf RSTACK
        return
;*********************************************************************************
;***************************************************
;!### CODE  Word: _rcall_return
__rcall_return
        ; Push base address of return table.
        movlw low __return_table
        call _push_byte_literal
        movlw high __return_table
        movwf THIGH
        ; Add index into table.
        movpf RSTACK, WREG            ; get return index
        decf RSP, F                   ; fixup return stack ptr
        addwf TLOW, F                 ; add low bytes
        btfsc ALUSTA, C               ; test carry
        incf THIGH, F                 ; add carry
        
        ; call __pmem_fetch             ; fetch the return address to the stack
        call _pVX_QZX_                ; its mangled name, needed for the stripper
        
        ; Restore the program counter.
        movfp THIGH, PCLATH           ; high byte address
        movfp STACK, THIGH
        incf SP, F                    
        movpf TLOW, WREG             ; low byte address
        movfp STACK, TLOW
        incf SP, F
        movwf PCL                     ; return
;*********************************************************************************
;***************************************************
;!### CODE  Word: _fetch_lit
__fetch_lit
        call _pushtop                 ; Make room on the stack.

        ; Index into table is in WREG
        ; Add the base of the literal table

        addlw low __literal_table
        movwf TLOW
        
        clrf THIGH, F
        movlw high __literal_table
        addwfc THIGH, F

        goto _pVX_QZX_                 ; P@  : get the literal value
;*********************************************************************************
;******************************************************************************    
; MACRO  ints_disable_mac
ints_disable_mac macro
        bsf CPUSTA, GLINTD              ; disable interrupts
        incf GLINT_LOCK, F              ; increment the lock count
        endm
;******************************************************************************
;******************************************************************************    
; MACRO  ints_enable_mac
ints_enable_mac macro
;        btfss CPUSTA, GLINTD
;        goto $ + 4              ; skip if interrupts are enabled already
        tstfsz GLINT_LOCK       ; don't decrement GLINT Lock if its already 0
        dcfsnz GLINT_LOCK, F    ; decrement the lock count
        bcf CPUSTA, GLINTD      ; then enable interrupts if count is zero
        endm
;******************************************************************************
    

