\ interupt.f
\ words to handle saving and restoring registers through an interrupt


\ This code is Copyright 1998, Peter Csapo. It is has been distributed free
\ of charge, and no claim is made to its accuracy, reliability or performance
\ in any way. 




\ Changes:
\         v0.9.4
\              - Using post auto increment/decrement to optimise pushing and 
\                popping the RSTACK
\         v0.7 - Several important registers saved before INTERRUPT_ENTER is called
\                so those instructions are emitted by the compiler for each interrupt.
\
\         v0.6 - ALUSTA, BSR is now saved in unbanked memory. GLINT_LOCK does not
\                need to be saved. Interrupts do not need to be disabled when
\                switching banks.
\


\ It takes two instructions to push or pop a byte from the stack. Since an
\ interrupt can occur between those two instructions, they are done in the
\ order  for push: 1. decrement the SP         for pop: 1. copy from [SP]
\                  2. copy the value to [SP]            2. increment SP
\
\ This ensures that the top stack item is not corrupted by the interrupt
\ handler.

\ The interrupt handler itself uses the post increment/decrement feature of
\ the cpu, since is assured that interrupts are disabled during its execution.


\ Also, it is imperative that the interrupts are disabled during the following
\ events: A. The indirect registers are used for other than SP and RSP.





code INTERRUPT_ENTER
; save the critical registers
; the RSTACK is in post auto increment mode
; ALUSTA, BSR  are saved before this routine is called (emitted by compiler).
; PCLATH is pushed before this routine is called (emitted by compiler).
; We must push WREG, XHIGH, XLOW, PRODH, PRODL
        movfp WREG, RSTACK
        movfp XHIGH, RSTACK
        movfp XLOW, RSTACK
        movfp PRODH, RSTACK
        bsf   ALUSTA, FS1               ; turn off auto increment
        movfp PRODL, RSTACK
        incf GLINT_LOCK, F              ; increment the lock count
        return
endcode

code INTERRUPT_EXIT
        clrf GLINT_LOCK, F              ; clear the lock count, retfie will enable ints
        gprbank BANKSTACKS              ; address stack
        bcf ALUSTA, FS0                 ; set for post auto-decrement
        bcf ALUSTA, FS1                 ;    //
        movpf RSTACK, PRODL
        movpf RSTACK, PRODH
        movpf RSTACK, XLOW
        movpf RSTACK, XHIGH
        movfp RSTACK, WREG
        movfp RSTACK, PCLATH            ; 
        movfp BSR_SAVE, BSR             ; restore BSR
        movfp ALUSTA_SAVE,  ALUSTA      ; restores flags
        retfie                          ; return from interrupt without touching PCLATH
endcode

