\ math16.f

\ This code is Copyright 1998, Peter Csapo. It is has been distributed free
\ of charge, and no claim is made to its accuracy, reliability or performance
\ in any way. 


\ Modified August 7, 1997  -- changed to big endian 


code save_temp_regs
        ; save temp registers
        incf RSP, F
        movfp tempA, RSTACK
        incf RSP, F
        movfp tempB, RSTACK
        incf RSP, F
        movfp tempC, RSTACK
        incf RSP, F
        movfp tempD, RSTACK
        return
endcode
unremember

code restore_temp_regs
        ; restore temp registers
        bcf ALUSTA, FS0                 ; set for post auto-decrement
        bcf ALUSTA, FS1                 ;    //
        movpf RSTACK, tempD
        movpf RSTACK, tempC
        movpf RSTACK, tempB
        movpf RSTACK, tempA
        bsf ALUSTA, FS1                 ; unset auto-decrement
        return
endcode
unremember

code mult_common
        ; common code for um* (unsigned) and m* signed multiplication
        ;
        ; the 4 byte result will go in tempA:tempD where tempD is the
        ; least significant byte
        call _save_temp_regs
        ;
        movpf STACK, XHIGH      ; store high byte in XHIGH
        incf SP, F              ;
        movpf STACK, XLOW       ; store low byte in XLOW
        movfp XLOW, WREG        ; load wreg with low byte
        mulwf TLOW              ; multiply top low byte with it
        movpf PRODL, tempD      ; store result
        movpf PRODH, tempC
        movfp XHIGH, WREG       ; load wreg with high byte it
        mulwf THIGH             ; multiply top high byte with it
        movpf PRODL, tempB      ; store result
        movpf PRODH, tempA      ;
        mulwf TLOW              ; multiply high byte with TLOW
        movfp PRODL, WREG       ; add low result to tempC
        addwf tempC, F
        movfp PRODH, WREG       ; add high result to tempB
        addwfc tempB, F         ;
        movlw 0x00
        addwfc tempA, F         ; add carry to tempA
        movfp XLOW, WREG        ; load it into WREG
        mulwf THIGH             ; multiply it with HIGH
        movfp PRODL, WREG       ; add result to tempC
        addwf tempC, F
        movfp PRODH, WREG       ; add high result to tempB
        addwfc tempB, F         ;
        movlw 0x00
        addwfc tempA, F         ; add carry to tempA
        return
        ;
_mult_cleanup
        ; most significant word goes into top of stack
        movfp tempB, TLOW
        movfp tempA, THIGH
        ; put least significant word onto second stack item
        movfp tempD, STACK
        decf SP, F
        movfp tempC, STACK      ; 
        ;
        goto _restore_temp_regs
        ; defer return
endcode
unremember

code um*
        ; unsigned 16 x 16 multiply routine
        ; leaves a double (4 bytes) on the stack
        ; with most significant word on top
        ;
        call _mult_common
        ;
        ;
        goto _mult_cleanup
        ; defer return
endcode

code m*
        ; signed 16 x 16 multiply routine
        ; leaves a double (4 bytes) on the stack
        ; with most significant word on top
        ;
        call _mult_common
        ;
        btfss XHIGH, 7          ; arg2 negative?
        goto _mmult_sign_top
        movfp TLOW, WREG        ; negative, subtract introduced 2s comp excess
        subwf tempB, F
        movfp THIGH, WREG
        subwfb tempA, F
_mmult_sign_top
        btfss THIGH, 7          ; arg1 negative?
        goto _mmult_finished
        movfp XLOW, WREG        ; negative, subtract introduced excess
        subwf tempB, F
        movfp XHIGH, WREG
        subwfb tempA, F
_mmult_finished
        goto _mult_cleanup
        ; defer return
endcode

code *
        ; signed/unsigned 16x16 multiply with a 16 bit result left on top
        ;
        ; save temp registers
        incf RSP, F
        movfp tempA, RSTACK
        incf RSP, F
        movfp tempB, RSTACK
        ; the 2 byte result will go in tempA:tempB where tempB is the
        ; least significant byte
        movpf STACK, XHIGH       ; put high byte in XHIGH
        incf SP, F
        movpf STACK, XLOW      ; put high byte in XLOW
        incf SP, F
        movfp XLOW, WREG        ; load wreg with low byte
        mulwf TLOW              ; multiply top low byte with it
        movpf PRODL, tempB      ; store result
        movpf PRODH, tempA
        movfp XHIGH, WREG       ; load WREG with high byte
        mulwf TLOW              ; multiply with TLOW
        movfp PRODL, WREG       ; add low result to tempA
        addwf tempA, F
        movfp XLOW, WREG        ; load low byte into WREG
        mulwf THIGH             ; multiply it with HIGH
        movfp PRODL, WREG       ; add result to tempA
        addwf tempA, F
        ; product goes into top of stack
        movfp tempB, TLOW
        movfp tempA, THIGH
        ; restore temp registers
        movpf RSTACK, tempB
        decf RSP, F
        movpf RSTACK, tempA
        decf RSP, F
        return
endcode

code div_common
; Main division loop begins here.
_UM_divLoop
        bcf ALUSTA, C           ; shift another dividend bit into the residue
        rlcf XLOW, F            
        rlcf XHIGH, F
        rlcf tempD, F
        rlcf tempC, F

; Check if the residue is less than the divisor

        ; check if we shifted out a 1
        ; if we did the residue must be 
        ; bigger than the divisor.
        btfsc ALUSTA, C
        goto _UM_subca          ; residue > divisor

        ; check by subtracting the high bytes
        movfp THIGH, WREG
        subwf tempC, W
        btfss ALUSTA, Z
        goto _UM_notz
        movfp TLOW, WREG        ; high bytes are equal, subtract low bytes
        subwf tempD, W
_UM_notz
        btfss ALUSTA, C
        goto _UM_nosub          ; residue <= divisor

; We know residue > divisor, so calculate the least residue
_UM_subca
        movfp TLOW,WREG         ; residue = residue - denominator
        subwf tempD, F
        movfp THIGH, WREG
        subwfb tempC, F
        bsf ALUSTA, C           ; setup a 1 to be rolled into quotient

; Update the quotient.
_UM_nosub
        rlcf tempB, F           ; roll 1 or 0 into quotient
        rlcf tempA, F
        decfsz STACK, F         ; decrement the count
        goto _UM_divLoop
        return
endcode

code um/mod
; 32/16=>16 unsigned division. check that this is the forth standard.
;  ( dividendlow dividendhigh divisor -- remainder quotient )

        ; Temporaries:
        ; dividend low word in XHIGH:XLOW
        ; divisor in THIGH:TLOW
        ; count is on the data stack
        ; quotient in tempA:tempB
        ; residue in tempC:tempD
        ; count in tempC


        call _save_temp_regs
        ;
        ; move dividend into tempC:tempD:XHIGH:XLOW
        movpf STACK, tempC      ; preload the residue high byte
        incf SP, F
        movpf STACK, tempD      ; preload the residue low byte
        incf SP, F
        movpf STACK, XHIGH       ; high word of the divisor
        incf SP, F
        movpf STACK, XLOW      ;     //
        ; STACK is now count
        clrf STACK, F
        bsf STACK, 4            ; count = 16
        clrf tempB, F
        clrf tempA, F

        call _div_common
        ;
        ; push remainder onto the stack
        movfp tempD, STACK
        decf SP, F
        movfp tempC, STACK
        ; put quotient onto stack top
        movfp tempB, TLOW
        movfp tempA, THIGH
        ;
        goto _restore_temp_regs
        ; defer return
endcode

: um/  um/mod nip ;

: dnegate >R dup if negate R> invert else R> negate then ;

: fm/mod   ( d n -- r q) \ d is the dividend (double)
                         \ n is the divisor, r = remainder, q = quotient (floored)
        over 0< if   \ Dividend is negative.
           dup 0< if  \ Divisor is negative.
              negate >R dnegate R>
              um/mod swap negate swap
           else   \ Divisor is positive.
              >R dnegate R@ um/mod
              over 0<> if \ Non zero remainder.
                 1+ negate  \ Floor quotient.
                 swap R> swap -  \ Complement the remainder.
                 swap
              else  \ Zero remainder.
                 R> drop  \ Drop divisor.
                 negate
              then
           then
        else  \ Dividend is positive.
           dup 0< if   \ Divisor is negative.
              negate dup >R
              um/mod 
              over 0<> if  \ Non zero remainder.
                 1+ negate   \ Floor quotient.
                 swap R> -  \ Complement remainder.
                 swap
              else
                 R> drop   \ Drop divisor.
                 negate    \ Negate quotient.
              then
           else  \ Divisor is positive.
              um/mod
           then
        then ;
                 
                 
                     

: s>d  ( n -- d)  \ Converts a single number to a double number.
   dup 0< ;


: */  ( u1 u2 u3 -- u1*u2/u3)
        >R m* R> fm/mod swap drop ;
        
: */mod 
        >R m* R> fm/mod ;       
        
        
: /
        >R s>d R> fm/mod swap drop ;   \ Drop the remainder.

: /mod 
        >R s>d R> fm/mod ;

: mod
        >R s>d R> fm/mod drop ;           \ Drop the quotient.


