\ Standard FORTH words. Strives to be ANS FORTH compatible as much as
\ possible.

\ This code is Copyright 1998, Peter Csapo. It is has been distributed free
\ of charge, and no claim is made to its accuracy, reliability or performance
\ in any way. 


\ NOTE: @ and ! words implicitly assume that THE STACKS BANK IS 0!!


\ CHANGES:
\ Dec 1, 1997
\       -- Changed @ and ! to set the bank according to the high bits of the address.
\       -- Added 2@ and 2!
\ August 22, 1997
\       -- Fixed up < for  proper signed behaviour and added u< 
\          for unsigned comparisons.
\ Aug   7, 1997
\       -- changed to big endian
\ July 15, 1997
\       -- reworked execute so that it doesn't corrupt PCLATH after it is loaded.
\ July 10, 1997
\       -- modified lshift and rshift to perform correctly if given 0 as parameter.
\
\ July 8, 1997
\       -- added C@ and C!
\
\ Modified: July 6, 1997
\ -- Fixed an insidious bug in the swap code. It indexed through the
\    stack with the SP, letting an interrupt handler corrupt the top
\    of the stack.
\ Modified: June 17, 1997
\  -- Reworked enable_ints and disable_ints because of a relocation of the 
\     GLINT_LOCK register.

code clrwdt
        clrwdt
        return
endcode

code execute
; Jump to the address on top of the stack.
; That routine will return to the caller of execute.
        movfp THIGH, PCLATH
        movfp TLOW, WREG
        drop                    ; pop the top (macro not subroutine)
                                ; if this pop sequence was a call
                                ; we would be corrupting PCLATH
        
        movwf PCL               ; make the computed jump
endcode

code >r
; move a value from the data stack to the return stack
        incf RSP, F             ; grow return stack
        movpf THIGH, RSTACK     ; copy first byte into return stack
        incf RSP, F
        movpf TLOW, RSTACK      ; copy high byte
        goto _poptop
endcode

code 2>r
_2toR
        incf RSP, F             ; grow return stack
        movfp STACK, RSTACK     ; put second low byte onto return stack
        incf SP, F              ; shrink data stack
        incf RSP, F             ; grow the return stack
        movfp STACK, RSTACK     ; put second high byte on return stack
        incf SP, F              ; shrink data stack
        incf RSP, F             ; grow the return stack
        movfp THIGH, RSTACK     ; put first high byte onto return stack
        incf RSP, F             ; grow return stack
        movfp TLOW, RSTACK      ; put first low byte onto return stack
        goto _poptop
endcode

code r>
; move a value from the return stack to the data stack
        decf SP, F              ;  //
        movfp TLOW, STACK      ; put high byte onto stack
        decf SP, F
        movfp THIGH, STACK
        movfp RSTACK, TLOW     ; copy top of return stack to top of data stack
        decf RSP, F             ; shrink the return stack
        movfp RSTACK, THIGH
        decf RSP, F
        return
endcode

code r@
; copy the top value on the return stack to the data stack
        decf SP, F              ;  //
        movfp TLOW, STACK       ; pushdown stack
        decf SP, F
        movfp THIGH, STACK
        movfp RSTACK, TLOW     ; copy top of return stack to top of data stack
        decf RSP, F
        movfp RSTACK, THIGH
        incf RSP, F             ; restore the RSP
        ; The following instruction is necessary because we must avoid
        ; indexing through the stacks with the stack pointers to preserve
        ; their integrity across interrupts.
        movpf TLOW, RSTACK     ; restore the high byte
        return
endcode

code dup
; duplicate the top stack item
        dup     ; the assembly macro "dup"
        return
endcode

code drop
; drop a word from the stack
        drop
        return
endcode

code invert
        comf  TLOW, F
        comf  THIGH, F
        return
endcode

code negate
        movfp TLOW, WREG
        negw TLOW, F            ; sets carry if necessary
        comf THIGH, F           ; complement high byte
        movlw 0x00
        addwfc THIGH, F         ; add one to high if carry
        return
endcode

code or
        movpf STACK, WREG       ; get high byte
        incf SP, F              ; point to low byte
        iorwf THIGH, F          ; or the high byte
        movpf STACK, WREG       ; get low byte
        incf SP, F              ; shrink stack
        iorwf TLOW, F           ; or the low byte
        return
endcode

code and
        movpf STACK, WREG       ; get high byte
        incf SP, F              ;
        andwf THIGH, F          ; and the high byte
        movpf STACK, WREG       ; get low byte
        incf SP, F              ;
        andwf TLOW, F          ; and the low byte
        return
endcode

code xor
        movpf STACK, WREG       ; get high byte
        incf SP, F              ;
        xorwf THIGH, F          ; xor the high byte
        movpf STACK, WREG       ; get low byte
        incf SP, F              ;
        xorwf TLOW, F           ; xor the low byte
        return
endcode

code 1+
        infsnz TLOW, F          ;
        incf THIGH, F           ;
        return
endcode

code 1-
        decf TLOW, F
        movlw 0x00
        subwfb THIGH, F
        return
endcode

code +
        movfp STACK, WREG       ; get the high byte
        incf SP, F              ;
        addwf THIGH, F          ; add high byte to THIGH
        movfp STACK, WREG       ; get low byte
        addwf TLOW, F           ; add low bytes
        btfsc ALUSTA, C         ; test carry
        incf THIGH, F           ; add carry
        incf SP, F              ;
        return
endcode

code -
        movfp THIGH, WREG       ; get the high byte
        movfp STACK, THIGH
        incf SP, F              ;
        subwf THIGH, F          ; subtract high bytes
        movfp TLOW, WREG        ; get low bytes
        movfp STACK, TLOW
        subwf TLOW, F           ; subtract low bytes
        btfss ALUSTA, C         ; test borrow bit
        decf THIGH, F           ; subtract borrow
        incf SP, F              ;
        return
endcode

code swap
        movpf THIGH, XHIGH      ; temp. store first
        movpf TLOW, XLOW        ;   \\
        movfp STACK, THIGH      ; move second into top
        incf SP, F              ; address second high byte
        movfp STACK, TLOW       ; move second into top
        movfp XLOW, STACK       ; move first into second
        decf SP, F              ; fixup stack pointer
        movfp XHIGH, STACK      ; move first into second
        return
endcode

code 2*
        bcf ALUSTA, C           ; clear carry
        rlcf TLOW, F            ; shift left
        rlcf THIGH, F
        return
endcode

code 2/
        bcf ALUSTA, C           ; clear carry
        btfsc THIGH, 7          ; test sign bit
        bsf ALUSTA, C           ; set carry
        rrcf THIGH, F           ; signed shift right
        rrcf TLOW, F
        return
endcode

code u2/
        bcf ALUSTA, C           ; clear carry
        rrcf THIGH, F           ; unsigned shift right
        rrcf TLOW, F
        return
endcode

code  8lshift
        movfp TLOW, THIGH
        clrf TLOW, F
        return
endcode

code 8rshift
        movfp THIGH, TLOW
        clrf THIGH, F
        return
endcode

code =
        ; leave true if top two items are equal
        movfp THIGH, WREG       ; move high byte of first stack value into W
        cpfseq STACK            ; compare first and second high bytes
        goto __eq_not_eq1       ; not equal
        movfp TLOW, WREG        ; is equal
        incf SP, F              ; point to low byte of second stack item
        cpfseq STACK            ; compare first and second low bytes
        goto __eq_not_eq2       ; not equal
        incf SP, F
        goto __settop            ; true and return
__eq_not_eq1
        incf SP, F
__eq_not_eq2
        incf SP, F
        goto __clrtop            ; false and return
endcode

code true
        call _pushtop
        goto __settop
endcode

code false
        call _pushtop
        goto __clrtop
endcode

code 0=
        movpf TLOW, WREG
        iorwf THIGH, W
        tstfsz WREG
        goto __clrtop
        goto __settop
endcode
        
code 0<
        btfss THIGH, 7          ; test the sign bit
        goto __clrtop            ; not set = positive
        goto __settop            ; set = negative
endcode

code sp@
        call _pushtop
        movfp SP, TLOW
        clrf THIGH, F
        return
endcode

: depth
[code]
        movlw STACKORG
        call _push_byte_literal
endcode
        SP@ - 2/ 
        2 - 
        ;
        
code rdepth
        call _pushtop
        clrf THIGH, F
        movfp RSP, TLOW
        movlw RSTACKORG         ; initial value
        subwf TLOW, F           ; subtract from TLOW
        rrncf TLOW, F           ; divide by 2 by shifting
        return
endcode

code disable_ints
        ints_disable_mac
        return
endcode

code enable_ints
        ints_enable_mac
        return
endcode

code !
        ; Interrupts must be disabled before calling this function
        ; because it modifies the RSP
        ;
        ints_disable_mac
        ; store
        movpf RSP, XLOW         ; store RSP
        movfp TLOW, FSR         ; load address
        swapf THIGH, W          ; put variable bank in WREG
        ; pop the value off the stack
        call _poptop
        ; store the variable
        iorwf BSR,F             ; set the bank (ASSUMES HIGH NIBBLE BSR WAS 0 !!)
        movpf THIGH, INDF       ; store high byte
        incf FSR, F             ; indirect address next byte
        movpf TLOW, INDF        ; store low byte
        gprbank BANKSTACKS      ; switch banks
        ; pop next stack value
        call _poptop
        movfp XLOW,  RSP        ; restore RSP
        goto _enable_ints
endcode

code c!
        ; Interrupts must be disabled before calling this function
        ; because it modifies the RSP
        ;
        ints_disable_mac
        ; store
        movpf RSP, XLOW         ; store RSP
        movfp TLOW, FSR         ; load address
        swapf THIGH, W          ; put variable bank in WREG
        ; pop the value off the stack
        call _poptop
        ; store the variable
        iorwf BSR,F             ; set the bank (ASSUMES HIGH NIBBLE BSR WAS 0 !!)
        movpf TLOW, INDF        ; store low byte
        gprbank BANKSTACKS      ; switch banks
        ; pop next stack value
        call _poptop
        movfp XLOW,  RSP        ; restore RSP
        goto _enable_ints
endcode

code @
        ; Interrupts must be disabled before calling this function
        ; because it modifies the RSP
        ;
        ints_disable_mac
        ; fetch
        movpf RSP, XLOW         ; save RSP
        movfp TLOW, FSR         ; load address
        swapf THIGH, W          ; put variable bank in WREG
        iorwf BSR,F             ; set the bank (ASSUMES HIGH NIBBLE BSR WAS 0 !!)
        movpf INDF, THIGH       ; fetch high byte
        incf FSR, F             ; address low byte
        movpf INDF, TLOW        ; fetch low byte
        gprbank BANKSTACKS
        movfp XLOW, RSP         ; restore RSP
        goto _enable_ints
endcode

code c@ 
        ; Interrupts must be disabled before calling this function
        ; because it modifies the RSP
        ;
        ints_disable_mac
        ; fetch
        movpf RSP, XLOW         ; save RSP
        movfp TLOW, FSR         ; load address
        swapf THIGH, W          ; put variable bank in WREG
        iorwf BSR,F             ; set the bank (ASSUMES HIGH NIBBLE BSR WAS 0 !!)
        movpf INDF, TLOW        ; fetch low byte
        gprbank BANKSTACKS
        movfp XLOW, RSP         ; restore RSP
        clrf THIGH, F           ; zero THIGH
        goto _enable_ints
endcode

code p@
__pmem_fetch
; ( addr -- program_memory_word )  fetch a word from program memory
; We must disable interrupts to be sure that this word is not
; re-entered by an interrupt handler routine.
        ints_disable_mac
        movfp THIGH, TBLPTRH
        movfp TLOW, TBLPTRL
        tablrd 0, 0, TLOW       ; dummy read to update the TBLATCH
        tlrd 0, TLOW            ; read the low byte
        tlrd 1, THIGH           ; read the high byte
        goto _enable_ints
endcode



code pick
        ; Interrupts must be disabled before calling this function
        ; because it modifies the SP
        ;
        call _disable_ints
        movpf SP, WREG          ; save the current stack pointer
        incf TLOW, F            ; ensure pick index > 0
__pickloop
        incf SP, F              ; point to the next stack item
        incf SP, F
        decfsz TLOW, F          ; decrement pick index
        goto __pickloop         ; loop if necessary
        movfp STACK, THIGH      ; copy the picked item to the top
        incf SP, F
        movfp STACK, TLOW
        movfp WREG, SP          ; restore the stack pointer
        goto _enable_ints
endcode


code u< 
__unsigned_lessthan
        movpf THIGH, XHIGH
        movpf TLOW, XLOW
        call _poptop
        movfp XLOW, WREG
        subwf TLOW, F
        movfp XHIGH, WREG
        subwfb THIGH, F
        btfsc ALUSTA, C         ; check carry flag
        goto __clrtop            ; false, not less than
        goto __settop            ; true, less than
endcode

: u> swap u< ;

code <  
; Signed comparison.
     ; The unsigned comparison gives the correct result unless the two numbers have the 
     ; opposite sign.
     movpf STACK, WREG          ; high byte of second cell
     xorwf THIGH, W             ; result in WREG
     btfsc WREG, 7              ; test the xored sign bits
     call _swap                 ; sign bits different
     goto __unsigned_lessthan   ;
endcode
             
       
: >     swap < ;


: <>    = 0= ;

: >=  < 0= ;

: <= > 0= ;

: rot
        >R swap R> swap ;

: over   ( x1 x2 -- x1 x2 x1 )
        swap dup >R
        swap R> ;


: I     R@ ;

: J R> R> R@  swap >R swap >R ;

code ?dup
        movfp THIGH, WREG       
        iorwf TLOW, W           ; test for 0
        btfss ALUSTA, Z         ; if not
        goto _pushtop           ; then make a copy of the stack
        return
endcode


: nip  ( X2 X1 -- X1)
        swap drop ;

: tuck  ( X2 X1 -- X1 X2 X1 )
        dup >R swap R> ;

: +!    tuck @ + swap ! ;

: 0<>
        0= 0= ;

: lshift
        ?dup if
           0 do
              2*
           loop 
        then ;

: rshift
        ?dup if
           0 do
              u2/
           loop 
        then ;

: odd   1 and ;

: even  odd 0= ;

: 2drop drop drop ;

: 2dup over over ;

: 2swap rot >R rot R> ;

: 2r> R> R> swap ;

: 2+ 1+ 1+ ;

: 2! ( d addr -- )
    dup >R ! R> 1+ 1+ ! ;

: 2@ ( addr -- d )
    dup >R 1+ 1+ @ R> @ ;

: abs
        dup 0< if negate then ;

: min 2dup < if drop else nip then ;
: max 2dup < if nip else drop then ; 

32 constant bl

: cells 2* ;
: chars ;



