.(
        romutil.f
        Utility serial EEPROM words.
)


\ This code is Copyright 1998, Peter Csapo. It is has been distributed free
\ of charge, and no claim is made to its accuracy, reliability or performance
\ in any way. 



\ Basic words to read and write from the eeprom. Admittedly these words are not
\ well factored, but there is a need to conserve the hardware return stack.


\ fload 24C164.f   \ 24LC164 serial eeprom.
\ We may have lost compatibility with the 24C164 series somewhere down the line
\ since I have made so many changes since I last tested it on that series.


fload 24C65.f      \ 24C65 serial eeprom.


\ Error Handling code.
:zzz  no_ack_rom!
           CR P." No Ack on store" 
           I2C_ERROR ;

:zzz no_ack_rom@
           CR P." No Ack on fetch" 
           I2C_ERROR ;
        
\ ___________________________________________________________________
\ Store low byte in eeprom.
: (Crom!)  ( c linear_addr -- )
        >RomAddr
        SendRomAddr
        8lshift   \ Ready the data.
        TXBYTE    \ Send the data.
        0<> if    \ NO ack?
           no_ack_rom!
        then
        BSTOP     \ Send a stop bit.
        ;
        
: rom_ctrl_read ( ctrl_byte -- )         
         BSTART
[code]
         bsf THIGH,0     ; Specify read mode.
endcode
         TXBYTE      \ Send the control byte.
         if      \ No ack?
            no_ack_rom@
         then
         ;

: (Crom@)   ( addr -- c1)  \ Fetch the byte at location addr in the eeprom.
         >RomAddr    \ Make it into an I2C address.
         dup >R \ Control byte.
         SendRomAddr
         R>      \ Control byte.
         rom_ctrl_read
         RXBYTE      \ Get the byte.
         NACK            
         BSTOP       \ Send a stop bit.
         ;
\ ___________________________________________________________________



\ Buffer the internal eeprom address pointer, so we can optimise
\ reads.

variable current_rom_addr

: CURR_ROM_ADDR@ current_rom_addr @ ;
: CURR_ROM_ADDR! current_rom_addr ! ;




\ Initialize the rom unit.
: EEPROM_INIT
        I2C_INIT    \ Initialise the I2C interface.
        
        \ Synchronize the eeprom internal address pointer, and
        \ CURR_ROM_ADDR.
        
        1 CURR_ROM_ADDR! \ Set the current rom address to 1.
        0 (Crom@) drop   \ Read memory location 0, this sets the rom's
                         \ address counter to 1.
        ;
         

\ Read rom at the current rom address.
: Crom+@  ( -- c) 
         CURR_ROM_ADDR@
         dup 1+ CURR_ROM_ADDR!  \ Increment the current address pointer.
         
         ( CURR_ROM_ADDR)
         >RomAddr nip  \ Make a control byte.
         rom_ctrl_read
         RXBYTE      \ Get the byte.
         NACK
         BSTOP       \ Send a stop bit.
         ;

\ Read a byte from the rom.
: Crom@  ( addr -- c)
        
        dup CURR_ROM_ADDR@ = if  \ Read next byte.
            drop Crom+@ exit
        then
        
        dup 1+ CURR_ROM_ADDR!
        (Crom@) 
        ;

        
\ Write a byte to the rom.                
: Crom! ( c addr -- )
       dup 1+ CURR_ROM_ADDR!
       (Crom!) ;

        
\ Fetch the word at location addr, uses sequential reads. Faster than making
\ two separate memory reads.
: rom@  ( addr -- n1)
        dup CURR_ROM_ADDR@ = if
            drop 
            Crom+@ Crom+@
        else 
            dup 2+ CURR_ROM_ADDR!
            >RomAddr
            dup >R SendRomAddr R>
            rom_ctrl_read
            RXBYTE         \ Get the low byte.
            ACK
            RXBYTE         \ Get high byte.
            NACK
            BSTOP
        then
        8lshift or
        ;         

\ Fetch n consecutive bytes from the rom to data memory.
\ UNTESTED!!!
: rom>mem ( romaddr  memaddr n )

        \ Make sure n is greater than 0.
        dup 0 = if drop 2drop exit then
        dup 1 = if drop swap crom@ swap c! exit then

        dup >R   \ Temp save a copy of n.
        rot ( memaddr n romaddr) 
        
        \ Set the current rom addr.
        dup CURR_ROM_ADDR@ =
        if 
           >RomAddr nip  \ Make a control byte.
        else
           dup CURR_ROM_ADDR!
           >RomAddr
           dup >R   \ Save control byte.
           SendRomAddr 
           R>       \ Restore control byte.
        then
        
        rom_ctrl_read \ Send read control byte.

        ( memaddr n )
         1- 0 do
           RXBYTE ACK
           over C!
           1+       \ Increment memory address.
         loop
         
         RXBYTE NACK BSTOP \ Get last byte.
         swap C!
         
         \ Set curr_rom_addr.
         CURR_ROM_ADDR@ R> + CURR_ROM_ADDR!
         ;
         
        
\ : 2rom@ ( addr -- d ) \ Read double word from rom.
\   dup rom@ swap 2+ rom@ ;
   
\ UNTESTED!!   
: 2rom@ ( addr -- d ) \ Read double word from rom.

        \ Set the rom address ptr.
        dup CURR_ROM_ADDR@ =
        if 
           >RomAddr nip  \ Make a control byte.
        else
           dup CURR_ROM_ADDR!
           >RomAddr
           dup >R   \ Save control byte.
           SendRomAddr 
           R>       \ Restore control byte.
        then
        
        rom_ctrl_read \ Send read control byte.

        RXBYTE ACK
        RXBYTE ACK
        8lshift or
        RXBYTE ACK
        RXBYTE NACK BSTOP
        8lshift or
        
        \ Set the current rom addr.
        CURR_ROM_ADDR@ 4 + CURR_ROM_ADDR!
        ;
        
\ Write a word to the rom.
: rom! ( n addr --) \ Write a 16 bit value to the rom.
        over over Crom!  \ Store the low byte.
        swap 8rshift swap 1+ Crom! \ Store the high byte.
        ;


: 2rom! ( d addr -- ) \ Write double value to rom.
        dup >R rot swap rom! R> 2+ rom! ;
 
