.( 
        serial.f
        Basic asynchronous serial port i/o routines.
)


\ This code is Copyright 1998, Peter Csapo. It is has been distributed free
\ of charge, and no claim is made to its accuracy, reliability or performance
\ in any way. 


code serial_enable
        periphbank BANK0
        bcf TXSTA, SYNC         ; set for async
        bsf RCSTA, SPEN         ; enable serial port
        return
endcode

code serial_disable
        periphbank BANK0
        bcf RCSTA, SPEN         ; disable serial port
        return
endcode

\ Set the baud rate determined by the SBRG byte and the clock speed.
( SBRG  --  )
code setBRG
        periphbank BANK0                ; address USART registers
        movfp TLOW, SPBRG               ; initialize baud rate generator
        goto _poptop
endcode


code tx_enable
        periphbank BANK0
        bsf TXSTA, TXEN                 ; enable output
        return
endcode

code tx_disable
        periphbank BANK0
        bcf TXSTA, TXEN                 ; disable output
        return
endcode

code tx_enable_parity
        periphbank BANK0
        bsf TXSTA, TX9
        return
endcode

code tx_disable_parity
        periphbank BANK0
        bcf TXSTA, TX9
        return
endcode

\ Generate even parity.
code gen_parity
        ; The input char is in TLOW and is not consumed.
        ; The parity bit is left in the low order bit of THIGH.
        ; This routine from AN547.        
        swapf TLOW, W
        xorwf TLOW, W
        movwf THIGH
        rrncf THIGH, F
        rrncf THIGH, F
        xorwf THIGH, W
        andlw 0x03
        addlw 0x01
        rrncf WREG, W
        movwf THIGH
        return
endcode

\ Output a byte, polling the trmt register to see if it is safe to send.
\ Assume that transmission is enabled.
code TX
; ( c --)
        periphbank BANK0
        
        ; Skip parity generation if only 8 bit transmissions.
        btfss TXSTA, TX9
        goto __SOUT_POLL_TRMT
        
        ; Generate the parity bit.
        call _gen_parity        ; puts parity bit in THIGH, 0

        
        ; poll TRMT to see if we can send
__SOUT_POLL_TRMT
        btfss TXSTA, TRMT
        goto __SOUT_POLL_TRMT

        ; Put the parity bit in TX9D
        bcf TXSTA, TX9D
        btfsc THIGH, 0
        bsf TXSTA, TX9D

        ; Send the byte.
        movfp TLOW, TXREG     
        goto _poptop
endcode


code rx_enable
        periphbank BANK0
        bsf RCSTA, CREN         ; continuous reception enable
        return
endcode

code rx_disable
        periphbank BANK0
        bcf RCSTA, CREN
        return
endcode

code rx_enable_parity
        periphbank BANK0
        bsf RCSTA, RX9
        return
endcode

code rx_disable_parity
        periphbank BANK0
        bcf RCSTA, RX9
        return
endcode

\ Input a byte. Assume that receive is enabled. Return -1 if a framing 
\ error or parity error occurred.
code RX
; ( -- c )
        ; poll RCIF till a byte is available
        periphbank BANK1
__RX_poll_RCIF
        btfss PIR, RCIF
        goto __RX_poll_RCIF
        
        ; Make space on the stack.
        call _pushtop
        
        periphbank BANK0
        
        ; Temp store the parity bit and the ferr bit into XLOW.
        movpf RCSTA, XLOW

        ; Read the data.        
        movpf RCREG, TLOW

        ; check for a framing error
        btfsc XLOW, FERR
        goto __settop            ; framing error return -1
        
        ; If 8 bit reception, skip parity check
        btfss RCSTA, RX9
        goto _RX_return
        
        ; check for a parity error
        call _gen_parity        ; generate the parity bit into THIGH
        movfp XLOW, WREG        ; retreive the parity bit
        xorwf THIGH, W          ; test parity bit against generated one
        btfsc WREG, 0
        goto __settop            ; parity not matched, return error.
        
        ; No errors.
_RX_return
        clrf THIGH, F
        return
endcode

code RX_overflow?
        ; (  -- f )
        periphbank BANK0
        btfss RCSTA, OERR
        goto _false
        goto _true
endcode
        

